function [x, j, res_norm, res_hist] = cgwsss(A, x, r, tol, maxiter)
%CGWSSS CGW method tailored for Shifted Skew-Symmetric systems
%
%  This function solves the shifted skew-symmetric matrix system A*x = b,
%  where A = sigma*I + S with nonzero sigma and S skew-symmetric, using
%  a version of the CGW algorithm designed specifically for SSS systems.
%  Author      :  Reijer Idema (http://www.reijeridema.nl)
%  Last edited :  15 September 2008
%
%  Parameters
%     A        :  coefficient matrix
%     x        :  initial solution x_0
%     r        :  initial residual r_0 = b - A*x_0
%     tol      :  stop criterion on the relative residual norm
%     maxiter  :  stop criterion on the number of iterations
%
%  Output
%     x        :  approximate solution to the system A*x = b
%     j        :  iteration count
%     res_norm :  residual norm
%     res_hist :  residual norm history through all iterations
%
%  Complexity
%     matrix-vector     :  1
%     innter products   :  2
%     vector updates    :  3
%
%  Memory requirement
%     matrix memory     :  1
%     vector memory     :  4 (not counting res_hist)

%  Note that CGW is more generally applicable than just for SSS systems,
%  however this version uses the special properties of a SSS coefficient
%  matrix with nonzero shift sigma, and thus only works for such systems.
%  Compared to the general version, as described in our research on
%  Krylov methods for SSS systems, this version eliminates z by working
%  with s_j = sigma*p_j and v_j = A*s_j = sigma*A*p_j.

%% Initialize variables
j = 0;
sigma = A(1,1);                                 % sigma should be nonzero

s = r;                                          % s_0 = sigma*p_0 = sigma*z_0 = r_0
rip = r'*r;                                     % residual inner product
norm_r0 = sqrt(rip);                            % norm_r0 = ||r_0||
res_norm = norm_r0;                             % res_norm = ||r_0||
res_hist = zeros(maxiter+1, 1);                 % residual history preallocation
res_hist(1) = res_norm;

%% MRS3 iteration loop
while (res_norm/norm_r0 > tol && j<maxiter)

   j = j+1;                                     % increase iteration counter
   v = A*s;                                     % v_j = A*s_j

   alpha = (sigma*rip)/(v'*r);                  % alpha = sigma*(r_j'*r_j)/(v_j'*r_j)
   x = x + (alpha/sigma)*s;                     % x_{j+1} = x_j + alpha*(s_j/sigma)
   r = r - (alpha/sigma)*v;                     % r_{j+1} = r_j - alpha*(v_j/sigma)

   rip_old = rip;
   rip = r'*r;                                  % rip = r_{j+1}'*r_{j+1}
   beta = -rip/rip_old;                         % beta = -(r_{j+1}'*r_{j+1})/(r_j'*r_j)
   s = r + beta*s;                              % s_{j+1} = r_{j+1} + beta*s_j

   res_norm = sqrt(rip);                        % residual norm
   res_hist(j+1) = res_norm;

end

%% End
