function [x, j, res_norm, res_hist] = gcr1(A, x, r, tol, maxiter)
%GCR1 Generalized Conjugate Residual method with the orthogonalization
%truncated after one step, also known as Orthomin(1).
%
%  This function solves the matrix system A*x = b, using the truncated
%  Generalized Conjugate Residual (GCR) method, or Orthomin(1).
%  For shifted skew-symmetric systems, i.e., A = sigma*I + S with
%  S skew-symmetric, this algorithm is equivalent to full GCR.
%  Author      :  Reijer Idema (http://www.reijeridema.nl)
%  Last edited :  15 September 2008
%
%  Parameters
%     A        :  coefficient matrix
%     x        :  initial solution x_0
%     r        :  initial residual r_0 = b - A*x_0
%     tol      :  stop criterion on the relative residual norm
%     maxiter  :  stop criterion on the number of iterations
%
%  Output
%     x        :  approximate solution to the system A*x = b
%     j        :  iteration count
%     res_norm :  residual norm
%     res_hist :  residual norm history through all iterations
%
% Complexity
%     matrix-vector     :  1
%     innter products   :  4
%     vector updates    :  4
%
% Memory requirement
%     matrix memory     :  1
%     vector memory     :  5 (not counting res_hist)

%  Note that there is a vector pair v_j and v_{j-1},
%  that need to be shifted in each iteration, i.e.,
%  create v_{j+1}, v_j=v_{j-1}, clear v_{j-1}, which is very slow.
%  For efficiency this vector pair has been implemented as a
%  matrix of the form V = [ v_A, v_B ].
%  In each iteration the oldest of the two matrix columns is updated,
%  while the other one is conserved.
%  To this purpose a set of indices it and io, that describe which column
%  is which, are kept and updated in each iteration.
%  In programming languages that support pointers, the same effect will be
%  easy to accomplish by working with 2 vectors v_A and v_B,
%  and swapping the pointers to them in each iteration.

%% Initialize variables
j = 0;
s = r;                                          % s_0 = r_0
beta = 1;                                       % beta_0 = 1

norm_r0 = norm(r,2);                            % norm_r0 = ||r_0||
res_norm = norm_r0;                             % res_norm = ||r_0||
res_hist = zeros(maxiter+1, 1);                 % residual history preallocation
res_hist(1) = res_norm;

it = 1; io = 2;                                 % index_this & index_other
dim = size(A,1);                                % problem dimension
V = [ zeros(dim,1) zeros(dim,1) ];              % V = [ v_-1 v_0 ]

%% MRS3 iteration loop
while (res_norm/norm_r0 > tol && j<maxiter)

   j = j+1;                                     % increase iteration counter
   t = it; it = io; io = t;                     % swap index_this & index_other
   V(:,it) = A*r;                               % v_j = A*r_{j-1}

   alpha = (V(:,it)'*V(:,io))/beta;             % alpha_j = v_j'*(v_{j-1)/beta_{j-1})
   s = r - (alpha/beta)*s;                      % s_j = r_{j-1} - alpha_j*(s_{j-1)/beta_{j-1})
   V(:,it) = V(:,it) - (alpha/beta)*V(:,io);    % v_j = v_j - alpha_j*(v_{j-1)/beta_{j-1})

   beta = norm(V(:,it),2);                      % beta_j = ||v_j||
   gamma = (r'*V(:,it))/beta;                   % gamma_j = r_{j-1}'*(v_j/beta_j)
   x = x + (gamma/beta)*s;                      % x_j = x_j + gamma_j*(s_j/beta_j)
   r = r - (gamma/beta)*V(:,it);                % r_j = r_j - gamma_j*(v_j/beta_j)

   res_norm = norm(r,2);                        % residual norm
   res_hist(j+1) = res_norm;

end

%% End
